from dataclasses import dataclass, field, make_dataclass
import typing


@dataclass(eq=True, frozen=True)
class SensorDescription:
    """
    data class for describing a snow sensor
    """

    code: str = "-1"  # code used within the applicable API
    name: str = "basename"  # desired name for the sensor
    description: str = None  # description of the sensor
    accumulated: bool = False  # whether the data is accumulated
    units: str = None  # Optional units kwarg
    extra: typing.Any = field(default=None, hash=False)  # Optional extra data for sub-class specific information


@dataclass(eq=True, frozen=True)
class InstrumentDescription(SensorDescription):
    """
    Extend the Sensor Description to include instrument
    """

    # description of the specific instrument for the variable
    instrument: str = None


class VariableBase:
    """
    Base class to store all variables for a specific datasource. Each
    datasource should implement the class. The goal is that the variables
    are synonymous across implementations.(i.e. PRECIPITATION should have the
    same meaning in each implementation).
    Additionally, variables with the same meaning should have the same
    `name` attribute of the SensorDescription. This way, if multiple datsources
    are used to sample the same variable, they can be written to the same
    column in a csv.

    Variables in this base class should ideally be implemented by all classes
    and cannot be directly used from the base class.
    """

    PRECIPITATION = SensorDescription()
    SWE = SensorDescription()
    SNOWDEPTH = SensorDescription()

    @staticmethod
    def _validate_sensor(sensor: SensorDescription):
        """
        Validate that a sensor is not using the default values since they
        are meaningless
        """
        default = SensorDescription()
        if sensor.name == default.name and sensor.code == default.code:
            raise ValueError(f"{sensor.name} is the default implementation")

    @classmethod
    def from_code(cls, code):
        """
        Get the correct sensor description from the code
        """
        for k, v in cls.__dict__.items():
            if isinstance(v, SensorDescription) and v.code == str(code):
                cls._validate_sensor(v)
                return v
        raise ValueError(f"Could not find sensor for code {code}")


class CdecStationVariables(VariableBase):
    """
    Available sensors from CDEC.
    Exhaustive list:
    http://cdec4gov.water.ca.gov/reportapp/javareports?name=SensList
    """

    PRECIPITATIONACCUM = SensorDescription("2", "ACCUMULATED PRECIPITATION", "PRECIPITATION, ACCUMULATED", False)
    PRECIPITATION = SensorDescription("45", "PRECIPITATION", "PRECIPITATION, INCREMENTAL", True)
    SNOWDEPTH = SensorDescription("18", "SNOWDEPTH", "SNOW DEPTH")
    SWE = SensorDescription("3", "SWE", "SNOW, WATER CONTENT", False)
    TEMP = SensorDescription("4", "AIR TEMP", "TEMPERATURE, AIR")
    TEMPAVG = SensorDescription("30", "AVG AIR TEMP", "TEMPERATURE, AIR AVERAGE")
    TEMPMIN = SensorDescription("32", "MIN AIR TEMP", "TEMPERATURE, AIR MINIMUM")
    TEMPMAX = SensorDescription("31", "MAX AIR TEMP", "TEMPERATURE, AIR MAXIMUM")
    RH = SensorDescription("12", "Relative Humidity", "RELATIVE HUMIDITY")
    # TODO confirm with CDWR if these depths are standard, no metadata available
    TEMPGROUND = SensorDescription("52", "GROUND TEMPERATURE INT", "GROUND TEMPERATURE SNOW/SOIL INTERFACE")
    TEMPGROUND25CM = SensorDescription("194", "GROUND TEMPERATURE -25CM", "GROUND TEMPERATURE OBS -25CM")
    TEMPGROUND50CM = SensorDescription("195", "GROUND TEMPERATURE -50CM", "GROUND TEMPERATURE OBS -50CM")
    TEMPGROUND100CM = SensorDescription("196", "GROUND TEMPERATURE -100CM", "GROUND TEMPERATURE OBS -100CM")
    SOLARRAD = SensorDescription("103", "SOLAR RADIATION", "SOLAR RADIATION")
    WINDSPEED = SensorDescription("9", "WIND SPEED", "WIND SPEED")
    WINDDIR = SensorDescription("10", "WIND DIRECTION", "WIND DIRECTION")


# Available sensors from Snotel
SnotelVariables = make_dataclass(
    "SnotelVariables",
    [
        (
            "SNOWDEPTH",
            SensorDescription,
            field(default=SensorDescription("SNWD", "SNOWDEPTH")),
        ),
        ("SWE", SensorDescription, field(default=SensorDescription("WTEQ", "SWE"))),
        (
            "TEMP",
            SensorDescription,
            field(default=SensorDescription("TOBS", "AIR TEMP")),
        ),
        (
            "TEMPAVG",
            SensorDescription,
            field(default=SensorDescription("TAVG", "AVG AIR TEMP", "AIR TEMPERATURE AVERAGE")),
        ),
        (
            "TEMPMIN",
            SensorDescription,
            field(default=SensorDescription("TMIN", "MIN AIR TEMP", "AIR TEMPERATURE MINIMUM")),
        ),
        (
            "TEMPMAX",
            SensorDescription,
            field(default=SensorDescription("TMAX", "MAX AIR TEMP", "AIR TEMPERATURE MAXIMUM")),
        ),
        (
            "PRECIPITATION",
            SensorDescription,
            field(default=SensorDescription("PRCPSA", "PRECIPITATION", "PRECIPITATION INCREMENT SNOW-ADJUSTED")),
        ),
        (
            "PRECIPITATIONACCUM",
            SensorDescription,
            field(default=SensorDescription("PREC", "ACCUMULATED PRECIPITATION", "PRECIPITATION ACCUMULATION")),
        ),
        # TODO for the SCAN network this appears to be "RHUM", we may need a new class
        (
            "RH",
            SensorDescription,
            field(default=SensorDescription("RHUMV", "Relative Humidity", "RELATIVE HUMIDITY")),
        ),
        (
            "STREAMVOLUMEOBS",
            SensorDescription,
            field(default=SensorDescription("SRVO", "STREAM VOLUME OBS", "STREAM VOLUME OBS")),
        ),
        (
            "STREAMVOLUMEADJ",
            SensorDescription,
            field(default=SensorDescription("SRVOX", "STREAM VOLUME ADJ", "STREAM VOLUME ADJ")),
        ),
    ]
    + [
        (
            f"TEMPGROUND{abs(d)}IN",
            SensorDescription,
            field(
                default=SensorDescription(
                    "STO",
                    f"GROUND TEMPERATURE -{d}IN",
                    f"GROUND TEMPERATURE OBS -{d}IN",
                    extra={"height_depth": {"value": -d, "unitCd": "in"}},
                )
            ),
        )
        for d in [2, 4, 8, 20]
    ]
    + [
        (
            f"SOILMOISTURE{abs(d)}IN",
            SensorDescription,
            field(
                default=SensorDescription(
                    "SMS",
                    f"SOIL MOISTURE -{d}IN",
                    f"SOIL MOISTURE PERCENT -{d}IN",
                    extra={"height_depth": {"value": -d, "unitCd": "in"}},
                )
            ),
        )
        for d in [2, 4, 8, 20]
    ]
    + [
        (
            f"TEMPPROFILENEG{abs(d)}IN" if d < 0 else f"TEMPPROFILE{d}IN",
            SensorDescription,
            field(
                default=SensorDescription(
                    "PTEMP",
                    f"PROFILE TEMPERATURE {d}IN",
                    f"PROFILE TEMPERATURE OBS{d}IN",
                    extra={"height_depth": {"value": d, "unitCd": "in"}},
                )
            ),
        )
        for d in [-8, 0, 8, 16, 24, 31, 39, 47, 55, 63, 71, 79, 87, 94, 102, 110, 118, 126]  # noqa: E501
    ],
    bases=(VariableBase,),
)


class MesowestVariables(VariableBase):
    """
    Available sensors from Mesowest
    There are a lot of variables here. Feel free to PR to add some
    https://developers.synopticdata.com/mesonet/v2/api-variables/
    """

    TEMP = SensorDescription("air_temp", "AIR TEMP")
    DEWPOINT = SensorDescription("dew_point_temperature", "DEW POINT TEMPERATURE")
    RH = SensorDescription("relative_humidity", "RELATIVE HUMIDITY")
    WINDSPEED = SensorDescription("wind_speed", "WIND SPEED")
    WINDDIRECTION = SensorDescription("wind_direction", "WIND DIRECTION")
    PRESSURE = SensorDescription("pressure", "PRESSURE")
    SNOWDEPTH = SensorDescription("snow_depth", "SNOWDEPTH")
    SOLARRADIATION = SensorDescription("solar_radiation", "SOLAR RADIATION")
    WETBULBTEMPERATURE = SensorDescription("wet_bulb_temperature", "WET BULB TEMPERATURE")
    SOILTEMP = SensorDescription("soil_temp", "SOIL TEMPERATURE")
    SOILTEMPIR = SensorDescription("soil_temp_ir", "SOIL TEMPERATURE IR")
    SWE = SensorDescription("snow_water_equiv", "SWE")
    NETSHORTWAVE = SensorDescription("net_radiation_sw", "NET SHORTWAVE RADIATION")
    NETLONGWAVE = SensorDescription("net_radiation_lw", "NET LONGWAVE RADIATION")
    STREAMFLOW = SensorDescription("stream_flow", "STREAMFLOW")


class USGSVariables(VariableBase):
    """
    To add more sensors:
    https://help.waterdata.usgs.gov/codes-and-parameters/parameters
    """

    DISCHARGE = SensorDescription("00060", "DISCHARGE", "DISCHARGE (CFS)")
    STREAMFLOW = SensorDescription("74082", "STREAMFLOW", "STREAMFLOW, DAILY VOLUME (AC-FT)")
    SNOWDEPTH = SensorDescription("72189", "SNOWDEPTH", "Snow depth, Meters")
    SWE = SensorDescription("72341", "SWE", "Water content of snow, millimeters")
    SOLARRADIATION = SensorDescription("72179", "SOLAR RADIATION", "Shortwave solar radiation, watts per square meter")
    UPSHORTWAVE = SensorDescription(
        "72185",
        "UPWARD SHORTWAVE RADIATION",
        "Shortwave radiation, upward intensity, watts per square meter",
    )
    DOWNSHORTWAVE = SensorDescription(
        "72186",
        "DOWNWARD SHORTWAVE RADIATION",
        "Shortwave radiation, downward intensity, watts per square meter",
    )
    NETSHORTWAVE = SensorDescription(
        "72201",
        "NET SHORTWAVE RADIATION",
        "Net incident shortwave radiation, watts per square meter",
    )
    NETLONGWAVE = SensorDescription(
        "72202",
        "NET LONGWAVE RADIATION",
        "Net emitted longwave radiation, watts per square meter",
    )
    DOWNLONGWAVE = SensorDescription(
        "72175",
        "DOWNWARD LONGWAVE RADIATION",
        "Longwave radiation, downward intensity, watts per square meter",
    )
    UPLONGWAVE = SensorDescription(
        "72174",
        "UPWARD LONGWAVE RADIATION",
        "Longwave radiation, upward intensity, watts per square meter",
    )
    SURFACETEMP = SensorDescription(
        "72405",
        "SURFACE TEMPERATURE",
        "Surface temperature, non-contact, degrees Celsius",
    )


class GeoSphereCurrentVariables(VariableBase):
    TEMP = SensorDescription("TL", "Air Temperature")
    SNOWDEPTH = SensorDescription("SCHNEE", "Snowdepth")
    PRECIPITATION = SensorDescription("RR", "Rainfall in the last 10 minutes", accumulated=True)
    TEMPGROUND10CM = SensorDescription("TB1", "Soil temperature at a depth of 10cm")
    TEMPGROUND20CM = SensorDescription("TB2", "Soil temperature at a depth of 20cm")
    TEMPGROUND50CM = SensorDescription("TB3", "Soil temperature at a depth of 50cm")


class GeoSphereHistVariables(VariableBase):
    """
    Variables that correspond to the DAILY historical Klima dataset

    Daily and hourly have different variable names
    https://dataset.api.hub.geosphere.at/v1/station/historical/klima-v1-1h/metadata
    https://dataset.api.hub.geosphere.at/v1/station/historical/klima-v1-1d/metadata
    """

    TEMP = SensorDescription("t7", "Air temperature 2m on observation date")
    SNOWDEPTH = SensorDescription("schnee", "Snowdepth")
    PRECIPITATION = SensorDescription("nied", "Precipitation Total", accumulated=True)


class CuesLevel1Variables(VariableBase):
    """
    Variables for CUES level1 data
    https://snow.ucsb.edu/index.php/query-db/

    Some variables report back with multiple instruments. See `UPSHORTWAVE`
    and `UPSHORTWAVE2` for two instrument specific implementations
    of the same variable.

    """

    TEMP = InstrumentDescription("air temperature", "AIR TEMP")
    RH = InstrumentDescription("RH", "RELATIVE HUMIDITY")
    LASERSNOWDEPTH = InstrumentDescription("laser snow depth", "LASER SNOWDEPTH")
    SNOWDEPTH = InstrumentDescription("snow depth", "SNOWDEPTH")
    NEWSNOWDEPTH = InstrumentDescription("new snow depth", "NEW SNOWDEPTH")
    SWE = InstrumentDescription("Snow Pillow (DWR) SWE", "SWE")
    # PRECIPITATION = InstrumentDescription(
    #     "nied", "Precipitation Total", accumulated=True
    # )
    TEMPSURFSNOW = InstrumentDescription("snow surface temperature", "SNOW SURFACE TEMPERATURE")
    DOWNSHORTWAVE = InstrumentDescription(
        "downward looking solar radiation",
        "DOWNWARD SHORTWAVE RADIATION",
    )
    UPSHORTWAVE = InstrumentDescription(
        "upward looking solar radiation",
        "UPWARD SHORTWAVE RADIATION",
        instrument="Eppley Lab precision spectral pyranometer",
    )
    UPSHORTWAVE2 = InstrumentDescription(
        "upward looking solar radiation",
        "UPWARD SHORTWAVE RADIATION 2",
        instrument="uplooking Sunshine pyranometer  direct and diffus",
    )
    DOWNSHORTWAVEIR = InstrumentDescription(
        "downward looking near-IR radiation",
        "DOWNWARD NIR SHORTWAVE RADIATION",
    )
    UPSHORTWAVEIR = InstrumentDescription(
        "upward looking near-IR radiation",
        "UPWARD NIR SHORTWAVE RADIATION",
    )


class MetNorwayVariables(VariableBase):
    """
    See https://frost.met.no/concepts2.html#calculationmethod
    for explanation of variable naming.
    All available variables are
    https://frost.met.no/elementtable
    """

    TEMP = SensorDescription(
        "air_temperature",
        "AIR TEMP",
        "Air temperature (default 2 m above ground), present value",
    )
    TEMPAVG = SensorDescription(
        "best_estimate_mean(air_temperature P1D)",
        "AVG AIR TEMP",
        "Homogenised daily mean temperature. The mean is an arithmetic mean of 24 hourly values (00-00 UTC),",
    )
    SNOWDEPTH = SensorDescription(
        "surface_snow_thickness",
        "SNOWDEPTH",
        "The depth of the snow is measured in cm from the ground to the top of"
        " the snow cover. (Code=-1 means 'no snow'"
        " and can be presented as '.')",
    )
    # SWE is only available as regional and interpolated datasets, so
    # metloom will NOT return data
    SWE = SensorDescription(
        "liquid_water_content_of_surface_snow",
        "SWE",
        "Snow water equivalent is a measure of the amount of water obtained"
        " if the snow is melted (as height in mm of a water column)",
    )
    PRECIPITATIONACCUM = SensorDescription(
        "accumulated(precipitation_amount)",
        "ACCUMULATED PRECIPITATION",
        "Total precipitation amount in gauge"
        " (accumulated since last emptying). Timing for emptying and"
        " algorithm for calculating the precipitation"
        " amount depends on sensortype",
    )
    PRECIPITATION = SensorDescription(
        "precipitation_amount",
        "PRECIPITATION",
        "Tipping bucket. The gauge tips for every 0.1 mm."
        " Each tip is registered along with the time stamp for the tip."
        " This is the basis for calcutation of precipitation sum per minute",
    )


class NWSForecastVariables(VariableBase):
    """
    See https://api.weather.gov/gridpoints/BOI/28,28
    for examples of variables
    """

    # Precipitation is not returned hourly
    PRECIPITATIONACCUM = SensorDescription("quantitativePrecipitation", "ACCUMULATED PRECIPITATION", accumulated=True)
    # PRECIPITATION = SensorDescription(
    #     "quantitativePrecipitation", "PRECIPITATION",
    #     accumulated=False
    # )
    TEMP = SensorDescription(
        "temperature",
        "AIR TEMP",
    )
    DEWPOINT = SensorDescription("dewpoint", "DEW POINT TEMPERATURE")
    RH = SensorDescription("relativeHumidity", "RELATIVE HUMIDITY")


class SnowExVariables(VariableBase):
    """
    Variables for SnowEx met stations data, refer to user guide for adding more
    variables

    Metadata:
    https://nsidc.org/sites/default/files/documents/user-guide/snex_met-v001-userguide.pdf
    """

    TEMP_20FT = SensorDescription(
        "AirTC_20ft_Avg",
        "AIR TEMP @20ft",
        units="deg C",
        description="Air temperature measured at 20 ft",
    )

    TEMP_10FT = SensorDescription(
        "AirTC_10ft_Avg",
        "AIR TEMP @10ft",
        units="deg C",
        description="Air temperature measured at 10 ft",
    )

    UPSHORTWAVE = SensorDescription(
        "SUp_Avg",
        "UPWARD SHORTWAVE RADIATION",
        units="w/m^2",
        description="Shortwave radiation measured with upward-facing sensor",
    )
    DOWNSHORTWAVE = SensorDescription(
        "SDn_Avg",
        "DOWNWARD SHORTWAVE RADIATION",
        description="Shortwave radiation measured with downward-facing sensor",
    )
    SNOWDEPTH = SensorDescription(
        "SnowDepthFilter(m)",
        "SNOWDEPTH",
        description="Snow surface height in meters w/ filtering",
    )

    TEMPGROUND5CM = SensorDescription("TC_5cm_Avg", "SOIL TEMP @ 5cm")
    TEMPGROUND20CM = SensorDescription("TC_20cm_Avg", "SOIL TEMP @ 20cm")
    TEMPGROUND50CM = SensorDescription("TC_50cm_Avg", "SOIL TEMP @ 50cm")


class CSASVariables(VariableBase):
    """
    Variable meta for the stations:
    SASP - https://snowstudies.org/wp-content/uploads/2026/01/SASP_Variable_Table.xlsx
    SBSP - https://snowstudies.org/wp-content/uploads/2026/01/SBSP_Variable_Table.xlsx
    PTSP - https://snowstudies.org/wp-content/uploads/2026/01/PTSP_Variable_Table.xlsx
    SGSB - https://snowstudies.org/wp-content/uploads/2023/11/SBSG_Variable_Table.xlsx
    """

    SNOWDEPTH = SensorDescription("Sno_Height_M", "SNOWDEPTH", units="meters")
    RH_LOWER = SensorDescription("Lo_RH", "RELATIVE HUMIDITY", units="%")
    RH_UPPER = SensorDescription("Up_RH", "RELATIVE HUMIDITY", units="%")

    STREAMFLOW_CFS = SensorDescription("Discharge_CFS", "STREAMFLOW", units="CFS")
    SURF_TEMP = SensorDescription(
        "Sno_IR_C",
        "SURFACE TEMP",
        units="deg C",
        description="Snow surface temperature",
    )
    UPPER_WINDSPEED = SensorDescription(
        "UpWind_Uavg_MS",
        "UPPER WIND SPEED",
        units="m/s",
        description="Wind speed at the upper location",
    )
    UPPER_WINDDIR = SensorDescription(
        "UpWind_Dir_Uavg",
        "UPPER WIND DIRECTION",
        units="degrees",
        description="Wind direction at the upper location",
    )
    LOWER_WINDSPEED = SensorDescription(
        "LoWind_Uavg_MS",
        "LOWER WIND SPEED",
        units="m/s",
        description="Wind speed at the lower location",
    )
    LOWER_WINDDIR = SensorDescription(
        "LoWind_Dir_Uavg",
        "LOWER WIND DIRECTION",
        units="degrees",
        description="Wind direction at the lower location",
    )
    DOWNWELLING_BROADBAND = SensorDescription(
        "PyUp_Unfilt_W",
        "DOWNWELLING BROADBAND RADIATION",
        units="w/m^2",
        description="Reflected Broadband radiation",
    )
    DOWNWELLING_NIR_SWIR = SensorDescription(
        "PyUp_Filt_W",
        "DOWNWELLING NIR/SWIR RADIATION",
        units="w/m^2",
        description="Reflected NIR/SWIR radiation",
    )
    UPWELLING_BROADBAND = SensorDescription(
        "PyDwn_Unfilt_W",
        "UPWELLING BROADBAND RADIATION",
        units="w/m^2",
        description="Incoming Broadband radiation",
    )

    UPWELLING_NIR_SWIR = SensorDescription(
        "PyDwn_Filt_W",
        "UPWELLING NIR/SWIR RADIATION",
        units="w/m^2",
        description="Incoming NIR/SWIR radiation",
    )

    PRECIPITATION = SensorDescription(
        "Day_H2O_mm",
        "DAILY PRECIP",
        accumulated=False,
        units="mm",
        description="Daily accumulated precipitation in mm",
    )

    TEMPGROUND = SensorDescription(
        "Soil_Surf_C",
        "GROUND TEMPERATURE",
        units="deg C",
        description="Temperature at soil interface",
    )
    TEMPGROUND10CM = SensorDescription(
        "Soil_10cm_C",
        "GROUND TEMPERATURE -10CM",
        units="deg C",
        description="Soil temperature at a depth of 10cm",
    )
    TEMPGROUND20CM = SensorDescription(
        "Soil_20cm_C",
        "GROUND TEMPERATURE -20CM",
        units="deg C",
        description="Soil temperature at a depth of 20cm",
    )
    TEMPGROUND40CM = SensorDescription(
        "Soil_40cm_C",
        "GROUND TEMPERATURE -40CM",
        units="deg C",
        description="Soil temperature at a depth of 40cm",
    )


class SAILStationVariables(VariableBase):
    """
    https://adc.arm.gov/discovery/#/results/site_code::guc
    """

    PRECIPITATION = SensorDescription(
        "accum_rtnrt",
        "PRECIPITATION",
        description=(
            "Accumulated amounts of precipitation over the sampling interval exceeding a threshold of 005mm "
            "or the accumulated amount of fine precipitation observed over the last hour"
        ),
        accumulated=False,  # not a cumulative sum, just "accumulated" over the sampling interval
        extra=dict(
            site="GUC",
            measurement="wbpluvio2",
            facility_code="M1",
            data_level="a1",
            units="mm",
        ),
    )
    SNOWDEPTH = SensorDescription(
        "snow_depth_intensity",
        "SNOWDEPTH",
        description="New snow height",
        accumulated=False,
        extra=dict(
            site="GUC",
            measurement="LD",
            facility_code="M1",
            data_level="b1",
            units="mm",
        ),
    )
    UP_BROADBAND = SensorDescription(
        "swup",
        "UP SHORTWAVE RADIATION",
        description="Surface upwelling shortwave hemispheric irradiance, hourly mean",
        accumulated=False,
        extra=dict(
            site="GUC",
            measurement="ARMBECLDRAD",
            facility_code="M1",
            data_level="c1",
            units="w/m^2",
        ),
    )
    DOWN_BROADBAND = SensorDescription(
        "swdn",
        "DOWN SHORTWAVE RADIATION",
        description="Best estimate of surface downwelling shortwave hemispheric irradiance, hourly mean",
        accumulated=False,
        extra=dict(
            site="GUC",
            measurement="ARMBECLDRAD",
            facility_code="M1",
            data_level="c1",
            units="w/m^2",
        )
    )



import logging
import pandas as pd
from metloom.pointdata import CSVPointData, StationInfo
from metloom.variables import SnowExVariables
import os


LOG = logging.getLogger(__name__)


class SnowExMetInfo(StationInfo):
    # Name, id, lat, long, elevation, http filename
    GM_STUDY_PLOT = ("Grand Mesa Study Plot", "GMSP", 39.05084, -108.06144, 10626,
                     "2017.06.21/SNEX_Met_GMSP2_final_output.csv")
    LS_OBS_SITE = ("Local Scale Observation Site", "LSOS", 39.05225, -108.09792, 9791,
                   "2016.10.09/SNEX_Met_LSOS_final_output.csv")
    MESA_EAST = ("Mesa East", "ME", 39.10358, -107.88383, 10105,
                 "2016.10.10/SNEX_Met_ME_final_output.csv")
    MESA_MIDDLE = ("Mesa Middle", "MM", 39.03954, -107.94174, 10286,
                   "2016.10.10/SNEX_Met_MM_final_output.csv")
    MESA_WEST = ("Mesa West", "MW", 39.03388, -108.21399, 9950,
                 "2016.10.09/SNEX_Met_MW_final_output.csv")


class SnowExMet(CSVPointData):
    """
    These data are stored in csv data formats
    """
    ALLOWED_VARIABLES = SnowExVariables
    ALLOWED_STATIONS = SnowExMetInfo

    # Data is in UTC
    UTC_OFFSET_HOURS = 0

    URL = "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX_Met.001/"
    DATASOURCE = "NSIDC"
    DOI = "https://doi.org/10.5067/497NQVJ0CBEX"

    def _file_urls(self, *args):
        return [os.path.join(self.URL, self._station_info.path)]

    def _assign_datetime(self, resp_df):
        resp_df['datetime'] = pd.to_datetime(resp_df['TIMESTAMP'])
        resp_df = resp_df.drop(columns=['TIMESTAMP']).set_index('datetime')
        return resp_df



import logging
from datetime import timedelta, timezone
from typing import List
import pandas as pd
import geopandas as gpd
import requests
from geopandas import GeoDataFrame

from metloom.dataframe_utils import merge_df, resample_whole_df
from metloom.pointdata.base import GenericPoint
from metloom.variables import SensorDescription, NWSForecastVariables


LOG = logging.getLogger(__name__)


class NWSForecastPointData(GenericPoint):
    """
    Implementation for NWS forecast API
    https://www.weather.gov/documentation/services-web-api

    We can call the points api to get the URL for the forecast, i.e.
    https://api.weather.gov/points/42,-119

    In this example, 3 forecast URLs are available

    "forecast": "https://api.weather.gov/gridpoints/BOI/28,28/forecast",
    "forecastHourly": "https://api.weather.gov/gridpoints/BOI/28,28/forecast/hourly",
    "forecastGridData": "https://api.weather.gov/gridpoints/BOI/28,28",

    forecast can be used to return the 12 hour increments

    forecastHourly returns hourly data (hourly in local tz)

    forecastGridData returns the 'raw' grid data (hourly UTC).

    We will use the `forecastGridData` endpoint for this implementation

    The API returns data from a grid, meaning the `geometry` column
    in the returned geodataframes will represent the **CENTER** of
    the forecast grid cell.

    """
    DATASOURCE = "NWS Forecast"
    ALLOWED_VARIABLES = NWSForecastVariables
    URL = "https://api.weather.gov"
    POINTS_FROM_GEOM_DEFAULTS = {
        'within_geometry': True,
        'token_json': "~/.frost_token.json",
        'buffer': 0.0
    }

    def __init__(
        self, station_id, name,
        initial_metadata=None, metadata=None,
    ):
        """
        Args:
            station_id: id of station
            name: name of station
            initial_metadata: shapely point required to find the
                forecast grid cell
            metadata: optional metadata for the station (shapely point)
        """
        if initial_metadata is None:
            raise ValueError("Initial metadata is required for the NWS Forecast class")
        super(NWSForecastPointData, self).__init__(
            station_id, name, metadata=metadata
        )
        self._inital_metadata = initial_metadata
        # default UTC time
        self._tzinfo = timezone(timedelta(hours=0))

        # set the forecast grid parameters
        self._gridx = None
        self._gridy = None
        self._office = None
        self._grid_outline = None

    def _get_initial_metadata(self):
        """
        Get all metadata from the API for one point.
        """
        # use the initial metadata to find the grid points for the forecast
        resp = requests.get(
            f"{self.URL}/points/{self._inital_metadata.y},{self._inital_metadata.x}"
        )
        resp.raise_for_status()
        data = resp.json()
        properties = data["properties"]

        return properties

    def _get_metadata(self):
        """
        See docstring for PointData._get_metadata
        This setts the _metadata and _gridx and _gridy parameters
        """
        properties = self._get_initial_metadata()
        self._gridx = properties["gridX"]
        self._gridy = properties["gridY"]
        self._office = properties["gridId"]

        # use the grid points to find the center of the forecast cell

        url = f"{self.URL}/gridpoints/" \
              f"{self._office}/{self._gridx},{self._gridy}"
        resp = requests.get(url)
        resp.raise_for_status()
        data = resp.json()
        grid_properties = data["properties"]
        # Parse the polygon into a geodataframe
        df_loc = gpd.GeoDataFrame.from_features([data])
        self._grid_outline = df_loc.geometry.values[0]
        # find the center
        center = df_loc.centroid[0]

        return gpd.points_from_xy(
            [center.x],
            [center.y],
            z=[grid_properties["elevation"]["value"] * 3.28084]  # convert to ft
        )[0]

    def _get_observations(self):
        """
        Get the hourly data for a 7 day forecast
        Example request: https://api.weather.gov/gridpoints/BOI/28,28
        """
        # ensure we have office, gridx, and gridy set
        if self._metadata is None:
            self._get_metadata()
        url = f"{self.URL}/gridpoints/" \
              f"{self._office}/{self._gridx},{self._gridy}"
        resp = requests.get(url)
        resp.raise_for_status()
        data = resp.json()
        return data["properties"]

    def _sensor_response_to_df(
        self, response_data, sensor, final_columns,
        resample_duration=None
    ):
        """
        Process the response from the API into a dataframe for 1 sensor

        Args:
            response_data: list of entries from the API
            sensor: single variable object
            final_columns: expected columns
            resample_duration: if a resample is desired, a duration that can
                be parsed by pandas

        Returns
            Geodataframe of data
        """
        # Get the list of response data
        sensor_data = response_data[sensor.code]
        unit_str = sensor_data["uom"].split("wmoUnit:")[-1]
        df = pd.DataFrame.from_records(sensor_data["values"])

        # Rename and resample
        column_map = {
            "validTime": "datetime",
            "value": sensor.name,
        }
        df.rename(
            columns=column_map,
            inplace=True,
        )

        # parse midway through dates
        # (example datetime is '2024-06-19T04:00:00+00:00/PT1H')
        date_starts = pd.to_datetime(
            df["datetime"].str.split("/").apply(lambda x: x[0])
        )
        date_durations = pd.to_timedelta(
            df["datetime"].str.split("/").apply(lambda x: x[1])
        )
        date_mids = date_starts + date_durations / 2.0
        df["datetime"] = date_mids
        df = df.set_index("datetime")

        # resample to the desired duration
        if resample_duration is not None:
            df = resample_whole_df(
                df, sensor,
                interval=resample_duration
            )

        # add other expected columns
        df[f"{sensor.name}_units"] = [unit_str] * len(df)
        df["site"] = [self.id] * len(df)

        # keep the column names
        final_columns += [
            sensor.name, f"{sensor.name}_units",
        ]

        df = GeoDataFrame(
            df, geometry=[self.metadata] * len(df)
        )

        # double check utc conversion
        df = df.tz_convert(self.desired_tzinfo)

        # set index so joining works
        df = df.filter(final_columns)
        df = df.loc[pd.notna(df[sensor.name])]
        return df

    def _get_data(
        self,
        variables: List[SensorDescription],
        desired_duration=None,
    ):
        """
        Args:
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
            desired_duration: desired resample duration ("D", "h"). Data is
                hourly be default
        Returns:
            GeoDataFrame of data, indexed on datetime, site
        """

        df = None
        final_columns = ["geometry", "site"]
        # Get data from the API
        response_data = self._get_observations()
        if response_data:
            # Parse data for each variable
            for sensor in variables:
                # TODO: how does resampling of non-hourly precip work
                sensor_df = self._sensor_response_to_df(
                    response_data, sensor, final_columns,
                    resample_duration=desired_duration
                )
                df = merge_df(df, sensor_df)

        if df is not None:
            if len(df.index) > 0:
                # Set the datasource
                df["datasource"] = [self.DATASOURCE] * len(df.index)
                df.reset_index(inplace=True)
                df.set_index(keys=["datetime", "site"], inplace=True)
                df.index.set_names(["datetime", "site"], inplace=True)
            else:
                df = None
        self.validate_sensor_df(df)
        return df

    def get_daily_forecast(
        self,
        variables: List[SensorDescription],
    ):
        """
        Get a geopandas dataframe with daily results for a 7 day forecast.
        The geometry column will be the center of the forecast gridcell

        Args:
            variables: list of variables to return
        """
        return self._get_data(variables, desired_duration="D")

    def get_hourly_forecast(
        self,
        variables: List[SensorDescription],
    ):
        """
        Get a geopandas dataframe with hourly results for a 7 day forecast.
        The geometry column will be the center of the forecast gridcell

        Args:
            variables: list of variables to return
        """
        return self._get_data(variables, desired_duration="h")

    def get_forecast(
        self,
        variables: List[SensorDescription],
    ):
        """
        Get a geopandas dataframe with hourly results for a 7 day forecast.
        The geometry column will be the center of the forecast gridcell

        Args:
            variables: list of variables to return
        """
        # Do not resample
        return self._get_data(variables)



"""
Location to keep readers that pull in a flat file like a csv.
"""
import os

import pandas as pd

from metloom.pointdata import PointData
from metloom.variables import SensorDescription
from enum import Enum
from pathlib import Path
from datetime import datetime, timezone, timedelta
import logging
import requests
from typing import List
import geopandas as gpd
import numpy as np


LOG = logging.getLogger(__name__)


class StationInfo(Enum):
    """
    Since there is not enough info via an API, csv readers rely on a list of stations
    and info associated with them to build the calls. This should allow for a common
    dataset with multiple stations to be verified and isolated
    """

    # Defined: Name, id, lat, long, http sub path to file

    @property
    def station_name(self):
        return self.value[0]

    @property
    def station_id(self):
        return self.value[1]

    @property
    def latitude(self):
        return self.value[2]

    @property
    def longitude(self):
        return self.value[3]

    @property
    def elevation(self):
        return self.value[4]

    @property
    def path(self):
        return Path(self.value[5])

    @classmethod
    def all_station_ids(cls):
        return [e.station_id for e in cls]

    @classmethod
    def all_points(cls):
        return [e.point for e in cls]

    @classmethod
    def from_station_id(cls, station_id):
        result = None
        for e in cls:
            if station_id == e.station_id:
                result = e
                break
        return result

    @property
    def point(self):
        return gpd.points_from_xy([self.longitude],
                                  [self.latitude],
                                  [self.elevation])[0]


class CSVPointData(PointData):
    """
    Some met station data is stored off in flat csv files. This class enables the
    management of downloading those files while still allowing a similar interface
    """
    ALLOWED_STATIONS = StationInfo
    UTC_OFFSET_HOURS = 0  # Allows users to specificy the timezone of the datasets

    def __init__(self, station_id, name=None, metadata=None, cache='./cache'):
        """
        See docstring for PointData.__init__
        """
        super(CSVPointData, self).__init__(
            station_id, name,
            metadata=metadata
        )
        self._raw_metadata = None
        self._tzinfo = timezone(timedelta(hours=self.UTC_OFFSET_HOURS))
        self._cache = Path(cache)
        self._station_info = None
        self.valid = False

    def _verify_station(self):
        """ Verifies the station is valid using the associated enum"""
        self._station_info = self.ALLOWED_STATIONS.from_station_id(self.id)

        if self._station_info is not None:
            # Auto assign name
            if self.name is None:
                self.name = self._station_info.station_name

            return True
        else:
            LOG.error(f"Station ID {self.id} is not valid, allowed id's are "
                      f"{', '.join(self.ALLOWED_STATIONS.all_station_ids())}")
            return False

    def _verify_sensor(self, resp_df, variable: SensorDescription):
        """ Verifies the station is valid using the associated enum"""
        if variable.code in resp_df.columns:
            return True
        else:
            LOG.debug(f"{variable.name} not found in {self.id} data")
            return False

    def _file_urls(self, *args):
        """Returns the url to the file containing the station data"""
        raise NotImplementedError('CSVPointData._file_urls() must be implemented to '
                                  'download csv station data.')

    def _assign_datetime(self, resp_df):
        raise NotImplementedError('CSVPointData._assign_datetime() must be implemented '
                                  'to download csv station data.')

    def _download(self, urls):
        """Download the file(s)"""
        filenames = []
        for url in urls:
            filename = self._cache.joinpath(Path(url).name)
            filenames.append(filename)
            if not filename.exists():
                with requests.get(url, stream=True) as r:
                    LOG.info(f'Downloading {Path(url).name}...')
                    lines = r.iter_lines()
                    with open(filename, mode='w+') as fp:
                        for line in lines:
                            fp.write(line.decode('utf-8') + '\n')
        return filenames

    def _get_one_variable(self, resp_df, period, variable: SensorDescription):
        """
        Retrieve a single variable and process it accordingly
        """
        method = "sum" if variable.accumulated else "average"
        if self._verify_sensor(resp_df, variable):
            isolated = resp_df[variable.code].loc[:]

            # TODO: This may only be true for SNOWEX
            isolated.loc[isolated == -9999] = np.nan
            if method == 'average':
                data = isolated.resample(period).mean()
            elif method == 'sum':
                data = isolated.resample(period).sum()
            else:
                raise Exception('Invalid aggregation method')
        else:
            data = None
            LOG.debug(f"No data returned for {variable.name}")

        return data

    def _get_data(self, start_date, end_date, variables: List[SensorDescription],
                  period):
        """
        Utilizes cached data or downloads the data
        """
        self.valid = self._verify_station()

        if not self.valid:
            return None

        urls = self._file_urls(self._station_info.station_id,
                               start_date, end_date)

        # Make the cache dir
        if not self._cache.is_dir():
            os.mkdir(self._cache)

        # Download data if it doesn't exist locally.
        files = self._download(urls)
        dfs = [pd.read_csv(f, index_col=False, low_memory=False) for f in files]
        resp_df = pd.concat(dfs)
        resp_df = self._assign_datetime(resp_df)

        # use a predefined index to show nans in the event of patch data
        df = pd.DataFrame(index=pd.date_range(start_date, end_date, freq=period,
                                              name='datetime'),
                          columns=[v.name for v in variables])

        # Use this instead .loc to avoid index on patchy data
        ind = (resp_df.index >= pd.Timestamp(start_date)) & (
            resp_df.index < pd.Timestamp(end_date)
        )
        isolated = resp_df.loc[ind, resp_df.columns]
        for i, variable in enumerate(variables):
            df_var = self._get_one_variable(isolated, period, variable)
            if df_var is not None:
                if not np.all(df_var.isnull()):
                    df.loc[df_var.index, variable.name] = df_var
                    df.loc[:, f"{variable.name}_units"] = variable.units
            else:
                df = df.drop(columns=[variable.name])

        # All nan data suggests no matching data
        if np.all(df.isnull()):
            return None

        # Make this a geodataframe
        df["site"] = [self.id] * len(df)
        df["datasource"] = [self.DATASOURCE] * len(df)
        # Make this a geodataframe
        df = gpd.GeoDataFrame(df, geometry=[self.metadata] * len(df))
        df = df.reset_index().set_index(["datetime", "site"])

        self.validate_sensor_df(df)
        return df

    def get_daily_data(self, start_date: datetime, end_date: datetime,
                       variables: List[SensorDescription]):
        return self._get_data(
            start_date, end_date, variables, "D"
        )

    def get_hourly_data(self, start_date: datetime, end_date: datetime,
                        variables: List[SensorDescription]):
        return self._get_data(
            start_date, end_date, variables, "h"
        )

    def _get_metadata(self):
        return self._station_info.point

    @classmethod
    def points_from_geometry(cls, geometry: gpd.GeoDataFrame,
                             variables: List[SensorDescription], within_geometry=True,
                             buffer=0.0, **kwargs):
        # Avoid multiple polys and use a buffer if requested.
        projected_geom = geometry.dissolve().buffer(buffer).to_crs(4326)

        gdf = gpd.GeoDataFrame(geometry=cls.ALLOWED_STATIONS.all_points(), data=[],
                               crs=4326)
        # Use the exact geometry to filter, otherwise use the bounds of the poly
        if within_geometry:
            search_area = projected_geom.iloc[0]
        else:
            search_area = projected_geom.envelope.iloc[0]

        filtered_gdf = gdf[gdf.within(search_area)]
        return filtered_gdf



import json
from collections import OrderedDict
from datetime import datetime, date
from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

import geopandas as gpd
from metloom.pointdata.geosphere_austria import (
    GeoSphereCurrentPointData, GeoSphereHistPointData
)
from metloom.variables import (
    GeoSphereCurrentVariables, GeoSphereHistVariables
)
from tests.test_point_data import BasePointDataTest


TODAY = date.today()


class TestGeoSphereCurrentPointData(BasePointDataTest):
    DATA_DIR = Path(__file__).parent.joinpath("data")
    EXPECTED_DATETIMES = pd.date_range(
        TODAY.isoformat(), periods=3, freq='h', tz='UTC'
    )

    @pytest.fixture(scope="class")
    def shape_obj(self):
        fp = self.DATA_DIR.joinpath("austria_box.shp")
        return gpd.read_file(fp)

    def _meta_response(self, *args, **kwargs):
        """
        Mccall airport station metadata return
        """
        mock = MagicMock()

        url = args[0]

        if 'metadata' in url:
            with open(
                self.DATA_DIR.joinpath("geosphere_mocks/meta_mock.json")
            ) as fp:
                response = json.load(fp)

        else:
            raise ValueError('Invalid test url provided')

        mock.json.return_value = response
        return mock

    def mock_station_response(self, *args, **kwargs):
        """
        Build a sub hourly response for time series data
        """
        if "metadata" in args[0]:
            return self._meta_response(*args)
        else:
            t_values = pd.date_range(
                TODAY.isoformat(), periods=13, freq='10 min', tz='UTC'
            )
            obj = {
                'media_type': 'application/json', 'type': 'FeatureCollection',
                'version': 'v1',
                'timestamps': [
                    t.isoformat() for t in t_values
                ], 'features': [
                    {'type': 'Feature', 'geometry': {'type': 'Point',
                                                     'coordinates': [
                                                         16.35638888888889,
                                                         48.24861111111111]},
                     'properties': {'parameters': {
                         'TL': {'name': 'Lufttemperatur', 'unit': '°C',
                                'data': [1.0, 2.0, 1.0, 3.0, 2.9, 5.0, 0.0,
                                         0.3, 0.2, 0.5, 1.1, 1.1, 1.1]}},
                         'station': '11035'}}]}
            mock_obj = MagicMock()
            mock_obj.json.return_value = obj
            return mock_obj

    @pytest.fixture()
    def station(self):
        return GeoSphereCurrentPointData("11035", "Tester")

    @pytest.mark.parametrize(
        "dt, expected", [
            (datetime(2024, 1, 24), datetime(2023, 10, 24)),
            (datetime(2023, 6, 24), datetime(2023, 3, 24)),
            (datetime(2023, 3, 1), datetime(2022, 12, 1)),
            # Check 3 months ago accounts for february dates
            (datetime(2025, 5, 30), datetime(2025, 2, 28)),

        ]
    )
    def test_back_3_months(self, dt, expected, station):
        result = station._back_3_months(dt)
        assert expected == result

    @pytest.mark.parametrize('stid, long, lat, elev', [
        ("11035", 16.35638888888889, 48.24861111111111, 649.60632),
    ])
    def test_get_metadata(self, station, stid, long, lat, elev):
        with patch(
            "metloom.pointdata.geosphere_austria.requests"
        ) as mock_requests:
            mock_get = mock_requests.get
            mock_get.side_effect = self._meta_response
            expected = gpd.points_from_xy([long], [lat], z=[elev])[0]
            result = station.metadata == expected
            assert result

    def test_get_hourly_data_fails(
        self, station
    ):
        """
        Test that we fail with dates greater than 3 months old
        """
        with patch("metloom.pointdata.geosphere_austria.requests.get",
                   side_effect=self.mock_station_response):
            with pytest.raises(ValueError):
                station.get_hourly_data(
                    datetime(2021, 1, 1, 0),
                    datetime(2021, 1, 1, 2),
                    [GeoSphereCurrentVariables.TEMP],
                )

    @pytest.mark.parametrize('var, expected_values, expected_dates', [
        (
            GeoSphereCurrentVariables.TEMP,
            [2.4833333333333334, 0.5333333333333333, 1.1],
            [t.isoformat() for t in EXPECTED_DATETIMES]
        ),
    ])
    def test_get_hourly_data(
        self, station, var, expected_values, expected_dates
    ):
        # Patch in the made up response
        with patch("metloom.pointdata.geosphere_austria.requests.get",
                   side_effect=self.mock_station_response):
            df = station.get_hourly_data(
                datetime(TODAY.year, TODAY.month, TODAY.day, 0),
                datetime(TODAY.year, TODAY.month, TODAY.day, 2),
                [var],
            )

        dt = [pd.to_datetime(d) for d in expected_dates]
        shp_point = gpd.points_from_xy(
            [16.35638888888889],
            [48.24861111111111],
            z=[649.60632],
        )[0]

        expected = gpd.GeoDataFrame.from_dict(
            OrderedDict({
                'site': [station.id] * len(dt),
                var.name: expected_values,
                'datetime': dt,
                f'{var.name}_units': ['°C'] * len(dt),
                'geometry': [shp_point] * len(dt),
                'datasource': ["GEOSPHERE"] * len(dt),
            }),
            geometry=[shp_point] * len(dt))
        expected.set_index(keys=["datetime", "site"], inplace=True)
        pd.testing.assert_frame_equal(df, expected)

    @pytest.mark.parametrize('w_geom, expected_sid, buffer', [
        (False, ['11266', '11125', '11121', '11320', '11123'], 0.15),
        # Use just bounds of the shapefile
        (True, ['11266'], 0.0),  # Filter to within the shapefile
    ])
    def test_points_from_geometry(self, shape_obj, w_geom, expected_sid,
                                  buffer):
        with patch("metloom.pointdata.geosphere_austria.requests.get",
                   side_effect=self.mock_station_response):
            pnts = GeoSphereCurrentPointData.points_from_geometry(
                shape_obj,
                [GeoSphereCurrentVariables.TEMP],
                within_geometry=w_geom,
                buffer=buffer
            )

        df = pnts.to_dataframe()
        assert df['id'].values == pytest.approx(expected_sid)


class TestGeoSphereHistPointData(BasePointDataTest):
    DATA_DIR = Path(__file__).parent.joinpath("data")

    @pytest.fixture(scope="class")
    def shape_obj(self):
        fp = self.DATA_DIR.joinpath("austria_box.shp")
        return gpd.read_file(fp)

    def _meta_response(self, *args, **kwargs):
        """
        Mccall airport station metadata return
        """
        mock = MagicMock()

        url = args[0]

        if 'metadata' in url:
            with open(
                self.DATA_DIR.joinpath("geosphere_mocks/klima_mock.json")
            ) as fp:
                response = json.load(fp)

        else:
            raise ValueError('Invalid test url provided')

        mock.json.return_value = response
        return mock

    def mock_station_response(self, *args, **kwargs):
        """
        Build a sub hourly response for time series data
        """
        if "metadata" in args[0]:
            return self._meta_response(*args)
        else:
            obj = {
                "media_type": "application/json", "type": "FeatureCollection",
                "version": "v1", "timestamps": [
                    "2023-01-20T00:00+00:00", "2023-01-21T00:00+00:00",
                    "2023-01-22T00:00+00:00", "2023-01-23T00:00+00:00",
                    "2023-01-24T00:00+00:00", "2023-01-25T00:00+00:00"
                ],
                "features": [
                    {"type": "Feature", "geometry": {
                        "type": "Point", "coordinates": [11.700833, 47.5075]
                    }, "properties": {
                        "parameters": {
                            "schnee": {
                                "name": "Gesamtschneehöhe zum 07 Uhr"
                                        "MEZ Termin",
                                "unit": "cm",
                                "data": [3.0, 18.0, 22.0, 18.0, 18.0, 14.0]
                            }
                        }, "station": "8807"}}
                ]
            }
            mock_obj = MagicMock()
            mock_obj.json.return_value = obj
            return mock_obj

    @pytest.fixture()
    def station(self):
        return GeoSphereHistPointData("8807", "Tester2")

    @pytest.mark.parametrize('stid, long, lat, elev', [
        ("8807", 11.700833, 47.5075, 3074.14708),
    ])
    def test_get_metadata(self, station, stid, long, lat, elev):
        with patch(
            "metloom.pointdata.geosphere_austria.requests"
        ) as mock_requests:
            mock_get = mock_requests.get
            mock_get.side_effect = self._meta_response
            expected = gpd.points_from_xy([long], [lat], z=[elev])[0]
            result = station.metadata == expected
            assert result

    @pytest.mark.parametrize('var, expected_values, expected_dates', [
        (
            GeoSphereHistVariables.SNOWDEPTH,
            [3.0, 18.0, 22.0, 18.0, 18.0, 14.0],
            ['2023-01-20T00:00:00+00:00', '2023-01-21T00:00:00+00:00',
             '2023-01-22T00:00:00+00:00', '2023-01-23T00:00:00+00:00',
             '2023-01-24T00:00:00+00:00', '2023-01-25T00:00:00+00:00']
        ),
    ])
    def test_get_daily_data(
        self, station, var, expected_values, expected_dates
    ):
        # Patch in the made up response
        with patch("metloom.pointdata.geosphere_austria.requests.get",
                   side_effect=self.mock_station_response):
            df = station.get_daily_data(
                datetime(2023, 1, 20),
                datetime(2023, 1, 25),
                [var],
            )

        dt = [pd.to_datetime(d) for d in expected_dates]
        shp_point = gpd.points_from_xy(
            [11.700833],
            [47.5075],
            z=[3074.14708],
        )[0]

        expected = gpd.GeoDataFrame.from_dict(
            OrderedDict({
                'site': [station.id] * len(dt),
                var.name: expected_values,
                'datetime': dt,
                f'{var.name}_units': ['cm'] * len(dt),
                'geometry': [shp_point] * len(dt),
                'datasource': ["GEOSPHERE"] * len(dt),
            }),
            geometry=[shp_point] * len(dt))
        expected.set_index(keys=["datetime", "site"], inplace=True)
        pd.testing.assert_frame_equal(df, expected)

    @pytest.mark.parametrize('w_geom, expected_sid, buffer', [
        (False, [
            '8807', '11900', '11901', '11902', '11903', '11910', '8800',
            '8805', '8806', '42', '151'
        ], 0.05),
        # Use just bounds of the shapefile
        (True, [
            '8807', '11900', '11901', '11903', '11910', '8800', '8805', '8806',
            '42', '151'
        ], 0.0),  # Filter to within the shapefile
    ])
    def test_points_from_geometry(self, shape_obj, w_geom, expected_sid,
                                  buffer):
        with patch("metloom.pointdata.geosphere_austria.requests.get",
                   side_effect=self.mock_station_response):
            pnts = GeoSphereHistPointData.points_from_geometry(
                shape_obj,
                [GeoSphereCurrentVariables.TEMP],
                within_geometry=w_geom,
                buffer=buffer
            )

        result = pnts.to_dataframe()['id'].values
        assert result == pytest.approx(expected_sid)



from datetime import datetime
from os.path import join
from pathlib import Path
from unittest.mock import patch, MagicMock

import geopandas as gpd
import pandas as pd
import pytest

from metloom.pointdata import CuesLevel1
from metloom.variables import CuesLevel1Variables
from tests.test_point_data import BasePointDataTest

DATA_DIR = str(Path(__file__).parent.joinpath("data/cues_mocks"))


class TestCuesStation(BasePointDataTest):

    @pytest.fixture(scope="function")
    def station(self):
        return CuesLevel1(None, None)

    @pytest.fixture(scope="class")
    def expected_meta(self):
        return gpd.points_from_xy(
            [-119.029128], [37.643093], [9661]
        )[0]

    @pytest.fixture(scope="class")
    def daily_expected(self, expected_meta):
        df = gpd.GeoDataFrame.from_dict(
            [
                {
                    "datetime": pd.Timestamp("2020-03-15 08:00:00+0000", tz="UTC"),
                    "DOWNWARD SHORTWAVE RADIATION": 95.64,
                },
                {
                    "datetime": pd.Timestamp("2020-03-16 08:00:00+0000", tz="UTC"),
                    "DOWNWARD SHORTWAVE RADIATION": 86.87,
                },
                {
                    "datetime": pd.Timestamp("2020-03-17 08:00:00+0000", tz="UTC"),
                    "DOWNWARD SHORTWAVE RADIATION": 182.23,
                },

            ],
            geometry=[expected_meta] * 3,
        )
        df["DOWNWARD SHORTWAVE RADIATION_units"] = ["Watts/meter^2"] * len(df)
        df["site"] = ["CUES"] * len(df)
        df["datasource"] = ["UCSB CUES"] * len(df)
        # needed to reorder the columns for the pd testing compare
        df = df.filter(
            [
                "datetime",
                "geometry",
                "DOWNWARD SHORTWAVE RADIATION",
                "site",
                "DOWNWARD SHORTWAVE RADIATION_units",
                "datasource",
            ]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df

    @classmethod
    def get_url_response(cls, resp="daily"):
        if resp == 'daily':
            with open(join(DATA_DIR, "daily_response.txt")) as fp:
                data = fp.read()
        elif resp == 'hourly':
            with open(join(DATA_DIR, "hourly_response.txt")) as fp:
                data = fp.read()
        else:
            raise RuntimeError(f"{resp} is an unknown option")

        obj = MagicMock()
        obj.content = data.encode()
        return obj

    def test_get_metadata(self, station, expected_meta):
        assert expected_meta == station.metadata

    def test_get_daily_data(self, station, daily_expected):
        with patch("metloom.pointdata.cues.requests") as mock_requests:
            mock_requests.post.side_effect = [
                self.get_url_response(),
            ]
            response = station.get_daily_data(
                datetime(2020, 3, 15),
                datetime(2020, 3, 17),
                [CuesLevel1Variables.DOWNSHORTWAVE],
            )
        pd.testing.assert_frame_equal(
            response.sort_index(axis=1),
            daily_expected.sort_index(axis=1)
        )

    def test_get_hourly_data(self, station):
        """
        Test that we get hourly data correctly.
        This also uses the `UPSHORTWAVE` variable so we can test
        that the instrument specific implementation of variables is working.
        """
        with patch("metloom.pointdata.cues.requests") as mock_requests:
            mock_requests.post.side_effect = [
                self.get_url_response(resp="hourly"),
            ]
            resp = station.get_hourly_data(
                datetime(2020, 4, 1), datetime(2020, 4, 2),
                [CuesLevel1Variables.UPSHORTWAVE],
            )
        resp = resp.reset_index()
        assert resp["datetime"].values[0] == pd.to_datetime("2020-04-01 08")
        assert resp["datetime"].values[-1] == pd.to_datetime("2020-04-02 07")
        assert resp["UPWARD SHORTWAVE RADIATION"].values[0] == -9.78
        assert resp["UPWARD SHORTWAVE RADIATION"].values[-1] == -8.44
        assert all(resp["site"].values == "CUES")

    def test_points_from_geometry_failure(self, station):
        with pytest.raises(NotImplementedError):
            station.points_from_geometry(None, None)



import geopandas as gpd
from shapely.geometry import Polygon

coords = [[-119.8, 37.7], [-119.2, 37.7], [-119.5, 38.2], [-119.8, 37.7]]
df = gpd.GeoDataFrame.from_dict({},
                                geometry=gpd.GeoSeries(Polygon(coords),
                                                       crs=4326))
df.to_file('triangle.shp')



"""
Tests our snowex reader which is pulling down csv data from the nsidc
All mock data is cut to the first 2 weeks of january in 2017. 2018 for GMSP

"""
import shutil
import pytest
from datetime import datetime, timedelta
from pathlib import Path
import geopandas as gpd
from unittest.mock import patch

from metloom.variables import SnowExVariables
from metloom.pointdata import SnowExMet


DATA_DIR = str(Path(__file__).parent.joinpath("data/snowex_mocks"))


class TestSnowEx:
    def copy_file(self, urls):
        files = []
        for url in urls:
            file = Path(DATA_DIR).joinpath(Path(url).name)
            cache = Path(__file__).parent.joinpath('cache')
            shutil.copy(file, cache.joinpath(file.name))
            files.append(file)
        return files

    @pytest.fixture(scope='function')
    def cache_dir(self):
        """Cachae dir where data is being downloaded to"""
        cache = Path(__file__).parent.joinpath('cache')
        yield cache
        if cache.is_dir():
            shutil.rmtree(cache)

    @pytest.fixture(scope='function')
    def station(self, cache_dir, station_id):
        with patch.object(SnowExMet, '_download', new=self.copy_file):
            pnt = SnowExMet(station_id, cache=cache_dir)
            yield pnt

    @pytest.mark.parametrize('station_id, expected', [
        ('GMSP', 'Grand Mesa Study Plot')
    ])
    def test_station_name(self, station, station_id, expected):
        """ Check auto assignment of the name"""
        station._verify_station()
        assert station.name == expected

    @pytest.mark.parametrize('station_id, variable, expected_mean', [
        ('LSOS', SnowExVariables.TEMP_10FT, -7.07817),
        ('MW', SnowExVariables.SNOWDEPTH, 1.03503),
    ])
    def test_get_daily_data(self, station, station_id, variable, expected_mean):
        """ Check data pulling """
        df = station.get_daily_data(datetime(2017, 1, 1),
                                    datetime(2017, 1, 15), [variable])

        # Assert it's a daily timeseries
        assert df.index.get_level_values('datetime').inferred_freq == 'D'
        assert df[variable.name].mean() == pytest.approx(expected_mean, abs=1e-5)

    @pytest.mark.parametrize('station_id, variable, expected_mean', [
        # Test a couple stations with different variables
        ('LSOS', SnowExVariables.TEMP_10FT, -7.86599),
        ('MW', SnowExVariables.SNOWDEPTH, 0.98625),

    ])
    def test_get_hourly_data(self, station, station_id, variable, expected_mean):
        """ Check auto assignment of the name"""
        df = station.get_hourly_data(datetime(2017, 1, 1, 11),
                                     datetime(2017, 1, 10, 23),
                                     [variable])

        # Assert it's hourly timeseries
        assert df.index.get_level_values('datetime').inferred_freq.lower() == 'h'
        assert df[variable.name].mean() == pytest.approx(expected_mean, abs=1e-5)

    @pytest.mark.parametrize("station_id, variable, start", [
        # GMSP doesnt have radiation
        ('GMSP', SnowExVariables.UPSHORTWAVE, datetime(2018, 1, 1, 11)),
        # GMSP has no data in Jan - 2017
        ('GMSP', SnowExVariables.SNOWDEPTH, datetime(2017, 1, 1, 11))

    ])
    def test_get_daily_none(self, station, station_id, variable, start):
        """ Test when a station doesn't have something, its handled"""
        end = start + timedelta(days=1)
        df = station.get_daily_data(start, end, [variable])
        assert df is None

    @pytest.mark.parametrize('within_geom, buffer, expected_count', [
        # Inside the geom is only one station
        (True, 0.0, 1),
        # inside the bounds of the geom is 3 stations
        (False, 0.0, 3),
        # Test Buffer use
        (True, 0.1, 4)
    ])
    def test_within_geometry(self, within_geom, buffer, expected_count):
        """ Use the within geometry on downloaded stations """
        search_poly = gpd.read_file(Path(DATA_DIR).joinpath('gm_polygon.shp'))
        df = SnowExMet.points_from_geometry(search_poly,
                                            [SnowExVariables.TEMP_20FT],
                                            within_geometry=within_geom,
                                            buffer=buffer)
        assert len(df.index) == expected_count



from datetime import datetime, timezone, timedelta
from typing import List
import geopandas as gpd
import numpy as np
import pandas as pd
import requests
import logging
from io import StringIO
from bs4 import BeautifulSoup
from geopandas import GeoDataFrame

from .base import PointData
from ..variables import USGSVariables, SensorDescription
from ..dataframe_utils import merge_df, append_df, resample_whole_df

LOG = logging.getLogger(__name__)


class USGSPointData(PointData):
    """
    Implement PointData methods for USGS data source.

    APIs and help
        https://waterservices.usgs.gov/rest/DV-Service.html#Service
        https://waterservices.usgs.gov/rest/IV-Service.html#Service
        https://waterservices.usgs.gov/rest/Site-Test-Tool.html

    """

    ALLOWED_VARIABLES = USGSVariables
    USGS_URL = "https://waterservices.usgs.gov/nwis/"
    META_URL = USGS_URL + "site/"
    DATASOURCE = "USGS"

    def __init__(self, station_id, name, metadata=None, duration=None):
        """
        See docstring for PointData.__init__
        """
        super(USGSPointData, self).__init__(
            str(station_id), name, metadata=metadata
        )
        self._tzinfo = None
        self._raw_metadata = None
        self.duration = duration

    def _get_all_metadata(self):
        if self._raw_metadata is None:
            params = {
                "format": "rdb",
                "sites": self.id,
                "siteOutput": "expanded",
                "siteStatus": "all"
            }
            resp = self._get_url_response(self.META_URL, params=params, parse='text')

            if resp:
                df = pd.read_csv(
                    StringIO(resp), delimiter="\t", skip_blank_lines=True,
                    comment="#"
                )
                df.drop(df[df['agency_cd'] != "USGS"].index, inplace=True)

            self._raw_metadata = df

        return self._raw_metadata

    def _get_metadata(self):
        """
        Get metadata
        """
        raw_meta = self._get_all_metadata()
        data = gpd.GeoDataFrame(
            geometry=gpd.points_from_xy(
                raw_meta["dec_long_va"],
                raw_meta["dec_lat_va"],
                z=raw_meta["alt_va"],
            )
        )
        data = data.set_crs("EPSG:4269").to_crs("EPSG:4326")
        return data.iloc[0]["geometry"]

    @staticmethod
    def _check_dates(start_date: datetime, end_date: datetime):
        """
        Ensure that datetimes are date only, without hour and minute, and that
        end_date is later than start_date.

        Args:
            start_date: datetime
            end_date: datetime

        Returns:
            start_date: datetime, date only
            end_date: datetime, date only
        """
        if hasattr(start_date, 'hour'):
            start_date = start_date.date()
        if hasattr(end_date, 'hour'):
            end_date = end_date.date()

        if not end_date >= start_date:
            LOG.error(
                f" end_date '{end_date}' must be later than start_date '{start_date}'"
            )
            raise ValueError("end_date must be later than start_date")

        return start_date, end_date

    @property
    def tzinfo(self):
        if self._tzinfo is None:
            self._tzinfo = self._get_tzinfo()
        return self._tzinfo

    def _get_tzinfo(self):
        raw_meta = self._get_all_metadata()
        tz_abbrev = raw_meta.iloc[0]["tz_cd"]
        tz_map = {
            "PST": timedelta(hours=-8.0),
            "PDT": timedelta(hours=-7.0),
            "MST": timedelta(hours=-7.0),
            "MDT": timedelta(hours=-6.0),
            "CST": timedelta(hours=-6.0),
            "CDT": timedelta(hours=-5.0),
            "EST": timedelta(hours=-5.0),
            "EDT": timedelta(hours=-4.0),
        }
        default = timedelta(hours=0)
        return timezone(tz_map.get(tz_abbrev, default))

    def _data_request(self, params, duration="dv"):
        """
        Make request to USGS and return JSON

        Args:
            params: dictionary of request parameters
            duration: daily ("dv") or instantaneous ("iv") values
        Returns:
            data: dict of response values
        """

        data = []
        contains_data = True
        url = self.USGS_URL + duration + "/"
        resp = self._get_url_response(url, params=params, parse='json')

        if "value" not in resp:
            LOG.warning(" Empty response from request")
            contains_data = False

        if len(resp["value"]["timeSeries"]) < 1:
            LOG.warning(
                " Requested site, sensor(s), and date range resulted in no data "
                "returned."
            )
            contains_data = False

        if contains_data:
            data = resp["value"]["timeSeries"][0]

            if not data["values"][0]["value"]:
                LOG.warning(
                    " Requested site, sensor(s), and date range resulted in no data "
                    "returned."
                )
                data = []

        return data

    def _sensor_response_to_df(
        self, response_data, sensor, final_columns, site_id,
        resample_duration=None
    ):
        """
        Convert the response data from the API to a GeoDataFrame Format and map columns
        in the dataframe

        Args:
            response_data: JSON list response from API
            sensor: SensorDescription obj
            final_columns: List of columns used for filtering
            site_id: site id
            resample_duration: To resample the array. [None, 'D', 'H']
        Returns:
            GeoDataFrame
        """

        if "values" not in response_data:
            LOG.warning(" Response does not contain expected data")
            raise ValueError("Failed parsing response for data")

        no_data_value = response_data["variable"]["noDataValue"]
        units = response_data["variable"]["unit"]["unitCode"]

        sensor_df = gpd.GeoDataFrame.from_dict(
            response_data["values"][0]["value"],
            geometry=[self.metadata] * len(response_data["values"][0]["value"]),
        )

        if sensor_df.empty:
            LOG.warning(" Data request resulted in no data returned")
            return []

        sensor_df.replace(no_data_value, np.nan, inplace=True)
        sensor_df["site"] = site_id
        sensor_df[f"{sensor.name}_units"] = units

        # cast values to float
        sensor_df["value"] = sensor_df["value"].astype('float32')

        final_columns += [sensor.name, f"{sensor.name}_units"]
        column_map = {"dateTime": "datetime", "value": sensor.name}
        sensor_df.rename(columns=column_map, inplace=True)

        # Daily data doesn't have timezone built in
        if pd.to_datetime(sensor_df["datetime"])[0].tzinfo is None:
            sensor_df["datetime"] = pd.DatetimeIndex(
                pd.to_datetime(sensor_df["datetime"])
            )
            # We will handle the timezone after resample to avoid
            # the case where resampling resets values to 0 hour of the day
            # for daily data
            handle_tz_after_resample = True

        # convert hourly or instantaneous data to utc prior to resample
        else:
            sensor_df["datetime"] = pd.to_datetime(
                sensor_df["datetime"], utc=True
            )
            handle_tz_after_resample = False

        if resample_duration:
            sensor_df = resample_whole_df(
                sensor_df.set_index("datetime"), sensor,
                interval=resample_duration
            ).reset_index()
            sensor_df = GeoDataFrame(sensor_df, geometry=sensor_df["geometry"])

        if handle_tz_after_resample:
            # handle the timezone if we didn't already
            sensor_df["datetime"] = sensor_df["datetime"].apply(
                self._handle_df_tz
            )

        # set index so joining works
        sensor_df.set_index("datetime", inplace=True)
        sensor_df.index = sensor_df.index.tz_convert(self.desired_tzinfo)
        sensor_df = sensor_df.filter(final_columns)
        sensor_df = sensor_df.loc[pd.notna(sensor_df[sensor.name])]
        return sensor_df

    def _get_data_fallback(self, params, duration):
        """
        Fallback for different sample times than requested not yet implemented.

        Args:
            params: request params with or without dur_code
            duration: daily ("dv") or instantaneous ("iv")
        Returns:
            response_data: dict of url response
            duration: daily ("dv") or instantaneous ("iv")
        """
        response_data = self._data_request(params, duration)
        return response_data, duration

    def _get_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
        duration_list: List[str],
        resample_duration=None
    ):
        """
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
            duration_list: list of USGS duration code, "dv" or "iv"
            resample_duration: optional if we need to resample the data
        Returns:
            GeoDataFrame of data, indexed on datetime, site
        """
        if not isinstance(duration_list, list):
            raise ValueError("duration list must be a list")

        start_date, end_date = self._check_dates(start_date, end_date)

        params = {
            'startDT': start_date.isoformat(),
            'endDT': end_date.isoformat(),
            'sites': self.id,
            'format': 'json',
            'siteStatus': 'all'
        }

        df = None
        final_columns = ["geometry", "site"]

        for sensor in variables:
            params["parameterCd"] = sensor.code
            response_data = None
            for duration in duration_list:
                response_data, response_duration = self._get_data_fallback(
                    params, duration
                )
                # Break if we got data
                if response_data:
                    break

            # format the data if we have any
            if response_data:
                sensor_df = self._sensor_response_to_df(
                    response_data, sensor, final_columns, self.id,
                    resample_duration=resample_duration
                )
                df = merge_df(df, sensor_df)
                df[sensor.name] = df[sensor.name].astype(float)

        if df is not None:
            if len(df.index) > 0:
                # Set the datasource
                df["datasource"] = [self.DATASOURCE] * len(df.index)
                df.reset_index(inplace=True)
                df.set_index(keys=["datetime", "site"], inplace=True)
                df.index.set_names(["datetime", "site"], inplace=True)
            else:
                df = None
        self.validate_sensor_df(df)

        return df

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        """
        return self._get_data(
            start_date, end_date, variables, ["dv", "iv"],
            resample_duration="24H"
        )

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        return self._get_data(
            start_date, end_date, variables, ["iv"], resample_duration="h"
        )

    def get_instantaneous_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        USGS 'instantaneous' data, which is generally 15 minutes.
        """
        return self._get_data(start_date, end_date, variables, ["iv"])

    @staticmethod
    def _get_url_response(url, params=None, parse='text'):
        """
        Get url response from USGS

        Args:
            url: url with formed query parameters
            params: optional dict of additional url call parameters
            parse: parsing format, either 'text' or 'json'

        Returns:
            response: DataFrame with query results
        """
        result = []
        if params:
            resp = requests.get(url, params)
        else:
            resp = requests.get(url)

        if resp.status_code != 200:
            if resp.status_code == 404:
                error_msg = "Response 404"
                soup = BeautifulSoup(resp.content, 'html.parser')
                status = soup.find('h1')
                if status is not None:
                    error_msg = status.get_text()
            elif resp.status_code == 400:
                error_msg = "Response 400"
            else:
                resp.raise_for_status()
                error_msg = "Error with request url"

            LOG.error(f"No data: {error_msg}")

        else:
            if parse == 'json':
                result = resp.json()
            elif parse == 'text':
                result = resp.text
            else:
                raise ValueError(f"Format '{parse}' not supported")

        return result

    @classmethod
    def _station_sensor_search(
        cls, meta_url, bounds, sensor: SensorDescription, dur="dv,iv", buffer=0.0
    ):
        """
        Search for USGS stations within a bounding box for the given sensor description.

        Args:
            meta_url: base url for metadata
            bounds: dictionary of lat/long bounds with keys minx, miny, maxx, maxy
            sensor: SensorDescription object
            dur: duration ("dv" or "iv")
            buffer: float of lat/lon degrees, buffer for bounding box
        """
        result = []
        bounds = bounds.round(decimals=5)
        minx = f"{bounds['minx'] - buffer}"
        miny = f"{bounds['miny'] - buffer}"
        maxx = f"{bounds['maxx'] + buffer}"
        maxy = f"{bounds['maxy'] + buffer}"

        url = (
            f'{meta_url}?format=rdb&bBox={minx}%2C{miny}%2C{maxx}%2C'
            f'{maxy}&siteStatus=active&hasDataTypeCd={dur}&parameterCd={sensor.code}'
        )

        response = cls._get_url_response(url, parse='text')

        if len(response):
            result = pd.read_csv(
                StringIO(response), delimiter="\t", skip_blank_lines=True, comment="#"
            )
            result = result[result['agency_cd'] == "USGS"]
        return result

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        **kwargs
    ):
        """
        See docstring for PointData.points_from_geometry

        Args:
            geometry: GeoDataFrame for shapefile from gpd.read_file
            variables: List of SensorDescription
        Returns:
            PointDataCollection
        """
        # assign defaults
        kwargs = cls._add_default_kwargs(kwargs)

        # Assume station search result is in 4326
        projected_geom = geometry.to_crs(4326)
        bounds = projected_geom.bounds.iloc[0]
        search_df = None
        station_search_kwargs = {}

        for variable in variables:
            result_df = cls._station_sensor_search(
                cls.META_URL, bounds, variable, buffer=kwargs["buffer"],
                **station_search_kwargs
            )

            if len(result_df):
                result_df["index_id"] = result_df["site_no"]
                result_df.set_index("index_id", inplace=True)
                search_df = append_df(
                    search_df, result_df
                ).drop_duplicates(subset=['site_no'])

        # return empty collection if we didn't find any points
        if search_df is None:
            return cls.ITERATOR_CLASS([])
        gdf = gpd.GeoDataFrame(
            search_df,
            geometry=gpd.points_from_xy(
                search_df["dec_long_va"],
                search_df["dec_lat_va"],
                z=search_df["alt_va"],
            ),
        )

        if not (search_df["dec_coord_datum_cd"] == "NAD83").all():
            LOG.error("Projection assumption for USGS is incorrect."
                      " Not all points are NAD 83")
        # convert results from NAD83 to WGS 84
        gdf = gdf.set_crs("EPSG:4269").to_crs("EPSG:4326")
        # filter to points within shapefile
        if kwargs['within_geometry']:
            filtered_gdf = gdf[gdf.within(projected_geom.iloc[0]["geometry"])]
        else:
            filtered_gdf = gdf

        points = [
            cls(row[0], row[1], metadata=row[2])
            for row in zip(
                filtered_gdf.index,
                filtered_gdf["station_nm"],
                filtered_gdf["geometry"],
            )
        ]

        return cls.ITERATOR_CLASS([p for p in points])



from datetime import datetime, date
from typing import List

import geopandas as gpd
import numpy as np
import pandas as pd
import requests
import logging
from geopandas import GeoDataFrame

from .base import PointData
from ..variables import (
    GeoSphereCurrentVariables, SensorDescription, GeoSphereHistVariables
)
from ..dataframe_utils import merge_df, resample_whole_df, \
    shp_to_box

LOG = logging.getLogger("metloom.pointdata.geosphere_austria")

M_TO_FT = 3.28084


class GeoSpherePointDataBase(PointData):
    """
    Implement PointData methods for GeoSphere Austria data source
    API documentation here
    https://dataset.api.hub.geosphere.at/v1/docs/index.html
    https://dataset.api.hub.geosphere.at/v1/docs/user-guide/resource.html

    Datasets available here https://data.hub.geosphere.at/dataset/

    We could either use the verified klima-v1 data or the
    raw tawes-v1 data. Kilma has hourly and daily, taws is 10minute
    and most current
    """

    ALLOWED_VARIABLES = None
    URL = "https://dataset.api.hub.geosphere.at"
    DATASOURCE = "GEOSPHERE"
    META_EXTENSION = None

    def __init__(self, station_id, name, metadata=None):
        """
        See docstring for PointData.__init__
        """
        super(GeoSpherePointDataBase, self).__init__(
            station_id, name, metadata=metadata
        )
        self._raw_metadata = None
        self._tzinfo = None

    @classmethod
    def _retrieve_all_metadata(cls):
        """
        Get the metadata we can search through for stations. The assumption
        is that we ONLY WANT TAWES stations

        The endpoint returns a json object with both `parameters` and
        `stations`. Parameters maps to ALL VARIABLES and stations
        maps to ALL STATIONS
        """
        url = cls.URL + cls.META_EXTENSION
        resp = requests.get(url)
        resp.raise_for_status()
        obj = resp.json()["stations"]
        df = pd.DataFrame.from_dict(obj)
        return df

    def _get_all_metadata(self):
        """
        Get all the raw metadata for a station. This is a list of sensor
        descriptions for the station
        Returns:
            A list of dictionaries describing the sensors at a station
        """
        if self._raw_metadata is None:
            all_meta = self._retrieve_all_metadata()
            meta_df = all_meta[all_meta["id"] == self.id]
            if len(meta_df) == 0:
                raise RuntimeError(f"No matching metadata for {self.id}")
            self._raw_metadata = meta_df.to_dict(orient="records")[0]
        return self._raw_metadata

    def _get_metadata(self):
        """
        See docstring for PointData._get_metadata
        """
        data = self._get_all_metadata()

        # TODO: gridded coords are EPSG:4325, are these also?
        return gpd.points_from_xy(
            [data["lon"]],
            [data["lat"]],
            # Convert elevation to feet
            z=[data["altitude"] * M_TO_FT],
        )[0]

    def _data_request(self, params):
        """
        Make get request and return JSON
        Args:
            params: dictionary of request parameters
        Returns:
            dictionary of response values
        """
        raise NotImplementedError("Need to implement")

    def _handle_df_tz(self, val):
        """
        Covert one entry from a df from cls.TZINFO to UTC
        """
        if pd.isna(val):
            return val
        else:
            return val.tz_convert(self.desired_tzinfo)

    def _sensor_response_to_df(self, response_data, sensor, final_columns,
                               resample_duration=None):
        """
        Convert the response data from the API to a GeoDataFrame
        Format and map columns in the dataframe
        Args:
            response_data: JSON list response from API
            sensor: SensorDescription obj
            final_columns: List of columns used for filtering
            resample_duration: duration to resample to
        Returns:
            GeoDataFrame
        """
        dt_values = response_data["timestamps"]
        params = response_data["features"][0]["properties"]["parameters"]
        values = params[sensor.code]["data"]
        unit = params[sensor.code]["unit"]
        # Build the dataframe
        sensor_df = pd.DataFrame.from_dict(
            {
                "datetime": dt_values,
                sensor.name: values,
                f"{sensor.name}_units": [unit] * len(values),
                "site": [self.id] * len(values)
            }
        )
        if all(pd.isna(sensor_df[sensor.name].values)):
            return None
        sensor_df.loc[pd.isna(sensor_df[sensor.name])] = np.nan
        sensor_df = gpd.GeoDataFrame(
            sensor_df,
            geometry=[self.metadata] * len(values),
        )

        final_columns = [sensor.name, f"{sensor.name}_units"] + final_columns
        sensor_df["datetime"] = pd.to_datetime(sensor_df["datetime"])

        # resample if necessary
        if resample_duration:
            sensor_df = resample_whole_df(
                sensor_df.set_index("datetime"), sensor,
                interval=resample_duration
            ).reset_index()
            sensor_df = GeoDataFrame(sensor_df, geometry=sensor_df["geometry"])

        sensor_df["datetime"] = sensor_df["datetime"].apply(self._handle_df_tz)
        # set index so joining works
        sensor_df.set_index("datetime", inplace=True)
        sensor_df = sensor_df.filter(final_columns)
        sensor_df = sensor_df.loc[pd.notna(sensor_df[sensor.name])]
        return sensor_df

    def _get_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
        desired_duration: str,
    ):
        """

        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
            desired_duration: duration code ['D', 'h']
        Returns:
            GeoDataFrame of data, indexed on datetime, site
        """
        params = {
            "parameters": ",".join([v.code for v in variables]),
            "station_ids": self.id,
            "start": start_date.isoformat(),
            "end": end_date.isoformat(),
        }
        df = None
        final_columns = ["geometry", "site"]
        response_data = self._data_request(params)

        if response_data:
            for sensor in variables:
                sensor_df = self._sensor_response_to_df(
                    response_data, sensor, final_columns,
                    resample_duration=desired_duration
                )
                df = merge_df(df, sensor_df)

        if df is not None:
            if len(df.index) > 0:
                # Set the datasource
                df["datasource"] = [self.DATASOURCE] * len(df.index)
                df.reset_index(inplace=True)
                df.set_index(keys=["datetime", "site"], inplace=True)
                df.index.set_names(["datetime", "site"], inplace=True)
            else:
                df = None
        self.validate_sensor_df(df)
        return df

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        **kwargs
    ):
        """
        See docstring for PointData.points_from_geometry

        The Austria Geosphere API does not allow filtering by variable.
        As a result, we do not filter according to which points have specific
        variables. The function arguments allow variables to be passed in
        to keep consistency with the same function from other classes.

        Args:
            geometry: GeoDataFrame for shapefile from gpd.read_file
            variables: List of SensorDescription. NOT USED FOR THIS CLASS
            within_geometry: filter the points to within the shapefile
            instead of just the extents. Default True
            buffer: buffer added to search box,
            filter_to_active: filter to active stations

        Returns:
            PointDataCollection
        """
        # assign defaults
        kwargs = cls._add_default_kwargs(kwargs)

        # Assume station search result is in 4326
        projected_geom = geometry.to_crs("EPSG:4326")

        # add buffer to geometry
        search_geom = projected_geom.buffer(kwargs["buffer"])

        # get metadata for all stations
        all_meta = cls._retrieve_all_metadata()

        # return empty collection if we didn't find any points
        if all_meta is None:
            return cls.ITERATOR_CLASS([])

        # convert to a geodataframe
        gdf = gpd.GeoDataFrame(
            all_meta,
            geometry=gpd.points_from_xy(
                all_meta["lon"], all_meta["lat"],
                all_meta["altitude"] * M_TO_FT
            )
        )
        # TODO: is this correct?
        gdf = gdf.set_crs("EPSG:4326")

        # filter to points within shapefile
        if kwargs['within_geometry']:
            filtered_gdf = gdf[gdf.within(projected_geom.iloc[0]["geometry"])]
        # filter to the overall bounding box
        else:
            box_df = shp_to_box(search_geom)
            filtered_gdf = gdf[gdf.within(box_df.iloc[0]["geometry"])]

        # filter to active stations
        if kwargs["filter_to_active"]:
            filtered_gdf = filtered_gdf.loc[filtered_gdf["is_active"] == "true"]

        points = [
            cls(row[0], row[1], metadata=row[2])
            for row in zip(
                filtered_gdf["id"],
                filtered_gdf["name"],
                filtered_gdf["geometry"],
            )
        ]
        return cls.ITERATOR_CLASS(points)


class GeoSphereCurrentPointData(GeoSpherePointDataBase):
    """
    Implement PointData methods for GeoSphere Austria data source
    API documentation here
    https://dataset.api.hub.geosphere.at/v1/docs/index.html
    https://dataset.api.hub.geosphere.at/v1/docs/user-guide/resource.html

    Datasets available here https://data.hub.geosphere.at/dataset/

    We use tawes-v1 data which consists of data from the last 3 months
    in 10 minute increment
    """

    ALLOWED_VARIABLES = GeoSphereCurrentVariables
    URL = "https://dataset.api.hub.geosphere.at"
    DATASOURCE = "GEOSPHERE"
    META_EXTENSION = "/v1/station/current/tawes-v1-10min/metadata"

    def _data_request(self, params):
        """
        Make get request and return JSON
        Args:
            params: dictionary of request parameters
        Returns:
            dictionary of response values
        """
        url = self.URL + "/v1/station/historical/tawes-v1-10min"
        resp = requests.get(url, params=params)
        resp.raise_for_status()
        return resp.json()

    @staticmethod
    def _back_3_months(dt):
        data_valid_start = dt - pd.DateOffset(months=3)
        return data_valid_start

    def _validate_dates(self, end_date):
        """
        Validate that the dates will work

        Args:
            end_date: datetime object for the end of the request
        """
        data_valid_start = self._back_3_months(date.today())
        if pd.to_datetime(end_date) < pd.to_datetime(data_valid_start):
            raise ValueError(
                f"This datasource does not have data older than 3 months. We "
                f"cannot fetch data for dates before"
                f" {data_valid_start.isoformat()}"
            )

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        Example query:
        https://dataset.api.hub.geosphere.at/v1/station/current/
        tawes-v1-10min?parameters=TL&station_ids=11035
        """
        self._validate_dates(end_date)
        return self._get_data(start_date, end_date, variables, "D")

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_hourly_data
        """
        self._validate_dates(end_date)
        return self._get_data(start_date, end_date, variables, "h")


class GeoSphereHistPointData(GeoSpherePointDataBase):
    """
    Implement PointData methods for GeoSphere Austria data source
    API documentation here
    https://dataset.api.hub.geosphere.at/v1/docs/index.html
    https://dataset.api.hub.geosphere.at/v1/docs/user-guide/resource.html

    Datasets available here https://data.hub.geosphere.at/dataset/

    We use klima-v1-1d data which consists of historical daily data.
    There is historical hourly data, but the parameter names are different
    and as such this has not bee implemented
    """

    ALLOWED_VARIABLES = GeoSphereHistVariables
    URL = "https://dataset.api.hub.geosphere.at"
    DATASOURCE = "GEOSPHERE"
    META_EXTENSION = "/v1/station/historical/klima-v1-1d/metadata"

    def _data_request(self, params):
        """
        Make get request and return JSON
        Args:
            params: dictionary of request parameters
        Returns:
            dictionary of response values
        """
        url = self.URL + "/v1/station/historical/klima-v1-1d"
        resp = requests.get(url, params=params)
        resp.raise_for_status()
        return resp.json()

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        Example query:
        https://dataset.api.hub.geosphere.at/v1/station/historical/klima-v1-1d
        ?station_ids=11401&start=2023-04-12&end=2023-04-14&parameters=schnee
        """
        return self._get_data(start_date, end_date, variables, None)



import json
import logging
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import List

import geopandas as gpd
import pandas as pd
import requests
from geopandas import GeoDataFrame

from metloom.dataframe_utils import append_df, merge_df, resample_whole_df
from metloom.pointdata.base import PointData
from metloom.variables import SensorDescription, MetNorwayVariables


LOG = logging.getLogger(__name__)


class MetNorwayPointData(PointData):
    """
    Class for the Norway Frost API
    https://frost.met.no/index.html

    To create a user, go here https://frost.met.no/auth/requestCredentials.html

    Data is provided by MET Norway, see license for details
    https://www.met.no/en/free-meteorological-data

    Element (variable) information can be found here
    https://frost.met.no/elementtable

    Observations/AvailableTimeSeries/ can be used to find out what types elements
    are available for a station or time range - we can use this to filter

    The Sources endpoint returns metadata. It can be used to filter
    based on geometry and variables

    For this class we will use default levels and timeoffsets. See more
    info here https://frost.met.no/concepts2.html#level-offset-filter

    It is important to note that this class does NOT implement all
    of the functionality of the frost API. The frost documentation
    is extensive and worth looking through.

    Read more about data quality codes here
    https://frost.met.no/dataclarifications.html

    Read more about general concepts: https://frost.met.no/concepts2.html

    """
    DATASOURCE = "MET Norway"
    ALLOWED_VARIABLES = MetNorwayVariables
    URL = "https://frost.met.no/"
    POINTS_FROM_GEOM_DEFAULTS = {
        'within_geometry': True,
        'token_json': "~/.frost_token.json",
        'buffer': 0.0
    }

    def __init__(
        self, station_id, name, token_json="~/.frost_token.json",
        metadata=None,
    ):
        """
        Args:
            station_id: id of station
            name: name of station
            token_json: path to file with authentication information
            metadata: optional metadata for the station
        """
        super(MetNorwayPointData, self).__init__(
            station_id, name, metadata=metadata
        )
        self._token_path = token_json

        # track how long the token is valid
        self._token_expires = None
        self._auth_header = None

        # default UTC time
        self._tzinfo = timezone(timedelta(hours=0))

    @classmethod
    def _get_token(cls, token_json):
        """
        Get token for authorization
        Args:
            token_json: path to json with credentials
        Returns:
            (auth headers, timestamp of expire)
        """
        # read in credentials
        token_json = Path(token_json).expanduser().absolute()
        with open(token_json, "r") as fp:
            obj = json.load(fp)
            _client_id = obj["client_id"]
            _client_secret = obj["client_secret"]

        url = cls.URL + "auth/accessToken"
        params = {
            "client_id": _client_id,
            "client_secret": _client_secret,
            "grant_type": "client_credentials"
        }
        resp = requests.post(url, data=params)
        resp.raise_for_status()
        result = resp.json()
        # get the token
        token = result["access_token"]
        # set the time when it expires
        _token_expires = datetime.now() + timedelta(
            seconds=result["expires_in"]
        )
        return {
            "Authorization": f"Bearer {token}"
        }, _token_expires

    def _token_is_valid(self):
        """
        function to check if token is valid
        """
        if self._token_expires is None:
            return False
        else:
            return datetime.now() >= self._token_expires

    @property
    def auth_header(self):
        """
        Get the auth header and set the expiration time
        """
        # get a new header if we need to
        if self._auth_header is None or not self._token_is_valid():
            token, expires = self._get_token(self._token_path)
            # Save when the token expires
            self._token_expires = expires
            # set auth header
            self._auth_header = token
        return self._auth_header

    @classmethod
    def _get_sources(
        cls, token_json="~/.frost_token.json", ids=None, types="SensorSystem",
        elements=None, geometry=None, validtime=None, name=None,
    ):
        """
        Get metadata for the source entitites defined in the Frost API.
        Use the query parameters to filter the set of sources returned.
        Leave the query parameters blank to select all sources.

        Args:
            token_json: path to json file with credentials
            ids: The Frost API source ID(s) that you want metadata for.
                Enter a comma-separated list to select multiple sources.
                For sources of type SensorSystem or RegionDataset, the source
                ID must be of the form <prefix><int> where <prefix> is SN
                for SensorSystem and TR, NR, GR, or GF for RegionDataset.
                The integer following the prefix may contain wildcards,
                e.g. SN18*7* matches both SN18700 and SN18007.
            types: The type of Frost API source that you want metadata for.
                [SensorSystem, InterpolatedDataset, RegionalDataset]
            elements: If specified, only sources for which observations are
                available for all of these elements may be included in the
                result. Enter a comma-separated list of search filters.
            geometry: Get Frost API sources defined by a specified geometry.
                Geometries are specified as either nearest(POINT(...)) or
                POLYGON(...) using WKT; see the reference section on the
                Geometry Specification for documentation and examples.
                If the nearest() function is specified, the output will
                include the distance in kilometers from the reference point.
            validtime: If specified, only sources that have been, or still are,
                valid/applicable during some part of this interval may be
                included in the result. Specify <date>/<date>, <date>/now,
                <date>, or now, where <date> is of the form YYYY-MM-DD,
                e.g. 2017-03-06. The default is 'now', i.e. only currently
                valid/applicable sources are included.
            name: If specified, only sources whose 'name' attribute matches
                this search filter may be included in the result.
        """
        url = cls.URL + "sources/v0.jsonld"
        if geometry is not None:
            geo_info = str(geometry.iloc[0].geometry)
        else:
            geo_info = None

        params = dict(
            ids=ids, types=types, elements=elements, geometry=geo_info,
            validtime=validtime, name=name
        )
        auth_header, _ = cls._get_token(token_json)
        resp = requests.get(url, params=params, headers=auth_header)
        if resp.status_code == 404:
            if ids:
                raise RuntimeError(f"Could not find metadata for {ids}")
            else:
                # No point were found
                result = None
        else:
            resp.raise_for_status()
            result = resp.json()["data"]
        return result

    def _get_all_metadata(self):
        """
        Get all metadata from the API for one point
        """
        result = self._get_sources(
            ids=[self.id], token_json=self._token_path
        )
        if len(result) != 1:
            raise RuntimeError("No metadata returned")

        return result[0]

    def _get_metadata(self):
        """
        See docstring for PointData._get_metadata
        """
        data = self._get_all_metadata()
        location_data = data["geometry"]["coordinates"]

        return gpd.points_from_xy(
            [location_data[0]],
            [location_data[1]],
        )[0]

    def _get_observations(self, ids, start_date, end_date, variables_names):
        """
        Args:
            ids: list of station ids
            start_date: datetime start date
            end_date: datetime end date
            variables_names: list of element names
        """
        url = self.URL + "observations/v0.jsonld"
        params = {
            "sources": ids,
            "referencetime": f"{start_date.isoformat()}/{end_date.isoformat()}",
            "elements": variables_names,
            # Defaults https://frost.met.no/concepts2.html#level-offset-filter
            "timeoffsets": "default",
            "levels": "default"
        }
        resp = requests.get(url, params=params, headers=self.auth_header)
        # 412 means there was no data found
        if resp.status_code == 412:
            # we could use /observations/availableTimeSeries/v0
            # to check this first
            LOG.debug(f"No data found for {ids}, {variables_names}")
            result = None
        else:
            resp.raise_for_status()
            result = resp.json()["data"]
        return result

    @staticmethod
    def _time_info_to_observation_time(
        reference_time, time_offset, time_resolution, timeseries_id
    ):
        """
        Get the observation time from the time info of an observation
        https://frost.met.no/concepts2.html#relationshipreftime

        Args:
            reference_time: string reference time
            time_offset: string time offset (PT1H)
            time_resolution: string time resolution (PT12H)
        Returns:
            observation_time
        """
        reference_time = pd.to_datetime(reference_time)
        time_offset = pd.to_timedelta(time_offset)
        time_resolution = pd.to_timedelta(time_resolution)
        observation_time = (
            reference_time + time_offset + time_resolution * timeseries_id
        )
        return observation_time

    def _sensor_response_to_df(
        self, response_data, sensor, final_columns,
        resample_duration=None
    ):
        """
        Process the response from the API into a dataframe for 1 sensor

        Args:
            response_data: list of entries from the API
            sensor: single variable object
            final_columns: expected columns
            resample_duration: if a resample is desired, a duration that can
                be parsed by pandas

        Returns
            Geodataframe of data
        """
        records = []
        for obs in response_data:
            ref_time = obs["referenceTime"]
            # filter to the relevant responses
            relevant_obs = [
                o for o in obs["observations"] if o["elementId"] == sensor.code
            ]
            if len(relevant_obs) == 0:
                # skip no data
                continue
            if len(relevant_obs) > 1:
                raise RuntimeError("This case is not implemented")
            else:
                # this is our winner
                o = relevant_obs[0]
                observation_time = self._time_info_to_observation_time(
                    ref_time, o["timeOffset"], o["timeResolution"],
                    o["timeSeriesId"]
                )
                records.append({
                    "datetime": observation_time,
                    "site": self.id,
                    sensor.name: o["value"],
                    f"{sensor.name}_units": o["unit"],
                    "quality_code": o["qualityCode"]
                })
        # return None for no data
        if not records:
            return None

        # keep the column names
        final_columns += [
            sensor.name, f"{sensor.name}_units", "quality_code"
        ]

        # create df
        sensor_df = pd.DataFrame.from_records(records)
        frequency = pd.infer_freq(pd.DatetimeIndex(sensor_df["datetime"]))
        sensor_df = GeoDataFrame(
            sensor_df, geometry=[self.metadata] * len(sensor_df)
        ).set_index("datetime")

        # resample to the desired duration
        if frequency != resample_duration and resample_duration is not None:
            sensor_df = resample_whole_df(
                sensor_df, sensor,
                interval=resample_duration
            )
            # Overwrite quality code if we resampled it
            sensor_df["quality_code"] = ["resampled"] * len(sensor_df)
            sensor_df = GeoDataFrame(sensor_df, geometry=sensor_df["geometry"])

        # double check utc conversion
        sensor_df = sensor_df.tz_convert(self.desired_tzinfo)

        # set index so joining works
        sensor_df = sensor_df.filter(final_columns)
        sensor_df = sensor_df.loc[pd.notna(sensor_df[sensor.name])]
        return sensor_df

    def _get_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
        desired_duration=None,
    ):
        """
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
        Returns:
            GeoDataFrame of data, indexed on datetime, site
        """

        df = None
        final_columns = ["geometry", "site"]
        # Get data from the API
        response_data = self._get_observations(
            [self.id], start_date, end_date, [v.code for v in variables]
        )
        if response_data:
            # Parse data for each variable
            for sensor in variables:
                sensor_df = self._sensor_response_to_df(
                    response_data, sensor, final_columns,
                    resample_duration=desired_duration
                )
                df = merge_df(df, sensor_df)

        if df is not None:
            if len(df.index) > 0:
                # Set the datasource
                df["datasource"] = [self.DATASOURCE] * len(df.index)
                df.reset_index(inplace=True)
                df.set_index(keys=["datetime", "site"], inplace=True)
                df.index.set_names(["datetime", "site"], inplace=True)
            else:
                df = None
        self.validate_sensor_df(df)
        return df

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        """
        return self._get_data(
            start_date, end_date, variables, desired_duration="D"
        )

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_hourly_data
        """
        return self._get_data(
            start_date, end_date, variables, desired_duration="h"
        )

    def get_event_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        Get data in original frequency from API
        """
        return self._get_data(
            start_date, end_date, variables, desired_duration=None
        )

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        **kwargs
    ):
        """
        See docstring for PointData.points_from_geometry

        Args:
            geometry: GeoDataFrame for shapefile from gpd.read_file
            variables: List of SensorDescription
            within_geometry: filter the points to within the shapefile
                instead of just the extents. Default True
            buffer: buffer added to search box
            token_json: Path to the public token for the mesowest api
                        default = "~/.frost_token.json"

        Returns:
            PointDataCollection
        """
        kwargs = cls._add_default_kwargs(kwargs)

        token_json = kwargs['token_json']
        projected_geom = geometry.to_crs("EPSG:4326")
        # buffer the geometry
        buffer = kwargs["buffer"]
        projected_geom = gpd.GeoDataFrame(
            geometry=projected_geom.dissolve().buffer(buffer)
        )
        # Take the outer bounds if we are not within the geometry
        if not kwargs["within_geometry"]:
            projected_geom = gpd.GeoDataFrame(geometry=projected_geom.envelope)

        # Loop over each variable and create a set of points.
        # _get_sources returns only points that have ALL variables passed
        # in, so looping over allows us to not exclude any points
        points_df = pd.DataFrame()
        for v in variables:
            source_info = cls._get_sources(
                token_json=token_json, geometry=projected_geom,
                elements=[v.code]
            )
            if source_info is not None:
                df = pd.DataFrame.from_records(source_info)
            else:
                df = None
            # build our final list
            points_df = append_df(
                points_df, df
            ).drop_duplicates(subset=['id'])

        if len(points_df) > 0:
            gdf = gpd.GeoDataFrame(
                points_df,
                # We don't get elevation back
                geometry=gpd.points_from_xy(
                    [p['coordinates'][0] for p in points_df["geometry"].values],
                    [p['coordinates'][1] for p in points_df["geometry"].values],
                ),
            )
            points = [
                cls(
                    row[0], row[1],
                    metadata=row[2]
                )
                for row in zip(
                    gdf["id"],
                    gdf["name"],
                    gdf["geometry"],
                )
            ]
        else:
            points = []

        # return a points iterator
        return cls.ITERATOR_CLASS(points)
